@@ -114,13 +114,14 @@ EncodeString(nsIUnicodeEncoder *encoder, const nsAFlatString &str, nsACString &r
         goto end;
     }
     p[maxlen] = 0;
-    result = p;
+    result.Assign(p);
 
-    rv = encoder->Finish(p, &len);
+    len = sizeof(buf) - 1;
+    rv = encoder->Finish(buf, &len);
     if (NS_FAILED(rv))
         goto end;
-    p[len] = 0;
-    result += p;
+    buf[len] = 0;
+    result.Append(buf);
 
 end:
     encoder->Reset();
@@ -494,6 +495,8 @@ nsStandardURL::BuildNormalizedSpec(const char *spec)
             Substring(spec + mHost.mPos, spec + mHost.mPos + mHost.mLen);
         if (tempHost.FindChar('\0') != kNotFound)
             return NS_ERROR_MALFORMED_URI;  // null embedded in hostname
+        if (tempHost.FindChar(' ') != kNotFound)
+            return NS_ERROR_MALFORMED_URI;  // don't allow spaces in the hostname
         if ((useEncHost = NormalizeIDN(tempHost, encHost)))
             approxLen += encHost.Length();
         else
@@ -557,6 +560,11 @@ nsStandardURL::BuildNormalizedSpec(const char *spec)
             LOG(("adding leading slash to path\n"));
             leadingSlash = 1;
             buf[i++] = '/';
+            // basename must exist, even if empty (bugs 113508, 429347)
+            if (mBasename.mLen == -1) {
+                mBasename.mPos = i;
+                mBasename.mLen = 0;
+            }
         }
 
         // record corrected (file)path starting position
@@ -621,7 +629,7 @@ nsStandardURL::BuildNormalizedSpec(const char *spec)
 }
 
 PRBool
-nsStandardURL::SegmentIs(const URLSegment &seg, const char *val)
+nsStandardURL::SegmentIs(const URLSegment &seg, const char *val, PRBool ignoreCase)
 {
     // one or both may be null
     if (!val || mSpec.IsEmpty())
@@ -630,12 +638,16 @@ nsStandardURL::SegmentIs(const URLSegment &seg, const char *val)
         return PR_FALSE;
     // if the first |seg.mLen| chars of |val| match, then |val| must
     // also be null terminated at |seg.mLen|.
-    return !strncmp(mSpec.get() + seg.mPos, val, seg.mLen)
-        && (val[seg.mLen] == '\0');
+    if (ignoreCase)
+        return !PL_strncasecmp(mSpec.get() + seg.mPos, val, seg.mLen)
+            && (val[seg.mLen] == '\0');
+    else
+        return !strncmp(mSpec.get() + seg.mPos, val, seg.mLen)
+            && (val[seg.mLen] == '\0');
 }
 
 PRBool
-nsStandardURL::SegmentIs(const char* spec, const URLSegment &seg, const char *val)
+nsStandardURL::SegmentIs(const char* spec, const URLSegment &seg, const char *val, PRBool ignoreCase)
 {
     // one or both may be null
     if (!val || !spec)
@@ -644,18 +656,25 @@ nsStandardURL::SegmentIs(const char* spec, const URLSegment &seg, const char *va
         return PR_FALSE;
     // if the first |seg.mLen| chars of |val| match, then |val| must
     // also be null terminated at |seg.mLen|.
-    return !strncmp(spec + seg.mPos, val, seg.mLen)
-        && (val[seg.mLen] == '\0');
+    if (ignoreCase)
+        return !PL_strncasecmp(spec + seg.mPos, val, seg.mLen)
+            && (val[seg.mLen] == '\0');
+    else
+        return !strncmp(spec + seg.mPos, val, seg.mLen)
+            && (val[seg.mLen] == '\0');
 }
 
 PRBool
-nsStandardURL::SegmentIs(const URLSegment &seg1, const char *val, const URLSegment &seg2)
+nsStandardURL::SegmentIs(const URLSegment &seg1, const char *val, const URLSegment &seg2, PRBool ignoreCase)
 {
     if (seg1.mLen != seg2.mLen)
         return PR_FALSE;
     if (seg1.mLen == -1 || (!val && mSpec.IsEmpty()))
         return PR_TRUE; // both are empty
-    return !strncmp(mSpec.get() + seg1.mPos, val + seg2.mPos, seg1.mLen); 
+    if (ignoreCase)
+        return !PL_strncasecmp(mSpec.get() + seg1.mPos, val + seg2.mPos, seg1.mLen); 
+    else
+        return !strncmp(mSpec.get() + seg1.mPos, val + seg2.mPos, seg1.mLen); 
 }
 
 PRInt32
@@ -713,6 +732,10 @@ nsStandardURL::ParseURL(const char *spec, PRInt32 specLen)
                                      &mPort);
         if (NS_FAILED(rv)) return rv;
 
+        // Don't allow mPort to be set to this URI's default port
+        if (mPort == mDefaultPort)
+            mPort = -1;
+
         mUsername.mPos += mAuthority.mPos;
         mPassword.mPos += mAuthority.mPos;
         mHost.mPos += mAuthority.mPos;
@@ -1503,14 +1526,9 @@ nsStandardURL::Equals(nsIURI *unknownOther, PRBool *result)
     NS_ENSURE_ARG_POINTER(unknownOther);
     NS_PRECONDITION(result, "null pointer");
 
-    nsRefPtr<nsStandardURL> otherPtr;
+    nsRefPtr<nsStandardURL> other;
     nsresult rv = unknownOther->QueryInterface(kThisImplCID,
-                                               getter_AddRefs(otherPtr));
-
-    // Hack around issue with MSVC++ not allowing the nsDerivedSafe to access
-    // the private members and not doing the implicit conversion to a raw
-    // pointer.
-    nsStandardURL* other = otherPtr;
+                                               getter_AddRefs(other));
     if (NS_FAILED(rv)) {
         *result = PR_FALSE;
         return NS_OK;
@@ -1568,7 +1586,7 @@ nsStandardURL::Equals(nsIURI *unknownOther, PRBool *result)
         rv = other->EnsureFile();
         if (NS_FAILED(rv)) {
             LOG(("nsStandardURL::Equals [other=%p spec=%s] other failed to ensure file",
-                other, other->mSpec.get()));
+                 other.get(), other->mSpec.get()));
             return rv;
         }
         NS_ASSERTION(other->mFile, "EnsureFile() lied!");
@@ -1692,7 +1710,7 @@ nsStandardURL::Resolve(const nsACString &in, nsACString &out)
     if (scheme.mLen >= 0) {
         // add some flags to coalesceFlag if it is an ftp-url
         // need this later on when coalescing the resulting URL
-        if (SegmentIs(relpath, scheme, "ftp")) {
+        if (SegmentIs(relpath, scheme, "ftp", PR_TRUE)) {
             coalesceFlag = (netCoalesceFlags) (coalesceFlag 
                                         | NET_COALESCE_ALLOW_RELATIVE_ROOT
                                         | NET_COALESCE_DOUBLE_SLASH_IS_ROOT);
@@ -1700,7 +1718,7 @@ nsStandardURL::Resolve(const nsACString &in, nsACString &out)
         }
         // this URL appears to be absolute
         // but try to find out more
-        if (SegmentIs(mScheme,relpath,scheme)) {
+        if (SegmentIs(mScheme, relpath, scheme, PR_TRUE)) {
             // mScheme and Scheme are the same 
             // but this can still be relative
             if (nsCRT::strncmp(relpath + scheme.mPos + scheme.mLen,
@@ -2219,8 +2237,8 @@ nsStandardURL::SetRef(const nsACString &input)
         refLen = buf.Length();
     }
 
-    ReplaceSegment(mRef.mPos, mRef.mLen, ref, refLen);
-    mPath.mLen += (refLen - mRef.mLen);
+    PRInt32 shift = ReplaceSegment(mRef.mPos, mRef.mLen, ref, refLen);
+    mPath.mLen += shift;
     mRef.mLen = refLen;
     return NS_OK;
 }
